"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSURL = void 0;
const commons_1 = require("./commons");
const strings_1 = require("./strings");
const tserrors_1 = require("./tserrors");
const types_1 = require("./types");
class TSURL {
    static from(source, options) {
        if (!(0, commons_1.$ok)(options)) {
            options = {};
            const protocol = source.protocol.toLowerCase();
            const w3c = TSURL.StandardURLProtocols.has(source.protocol.toLowerCase());
            if (!w3c && protocol.length) {
                options.acceptedProtocols = [protocol];
            }
            if (!source.search.length) {
                options.refusesParameters = true;
            }
        }
        return TSURL.url(source.href, options);
    }
    static url(source, opts = {}) {
        const url = _normalizeURLString(source);
        const len = url.length;
        const protocol = _completeProtocol(_findProtocol(url));
        if (protocol instanceof tserrors_1.TSError) {
            if (!!opts.throwsError) {
                throw protocol;
            }
            return null;
        }
        const w3c = TSURL.StandardURLProtocols.has(protocol);
        if (!w3c) {
            if (!_validateProtocol(protocol, opts.acceptedProtocols)) {
                if (!!opts.throwsError) {
                    new tserrors_1.TSError(`TSURL: unknown protocol in '${url}'`);
                }
                return null;
            }
        }
        let p = protocol.length;
        if (url[p] !== '/' || url[p + 1] !== '/') {
            if (!!opts.throwsError) {
                new tserrors_1.TSError(`TSURL: unable to find // before hostname in url '${url}'`);
            }
            return null;
        }
        let c = '';
        p += 2;
        let hostendpos = p;
        while (hostendpos < len) {
            c = url[hostendpos];
            if (c === '?' || c === '#' || c === '/') {
                break;
            }
            hostendpos++;
        }
        const atpos = url.lastIndexOf('@', hostendpos < len ? hostendpos : p);
        let auth = '';
        if (atpos >= 0) {
            auth = decodeURIComponent(url.slice(p, atpos));
            p = atpos + 1;
        }
        hostendpos = p;
        while (hostendpos < len) {
            if (TSURL.EndhostCharSet.has(url[hostendpos])) {
                break;
            }
            hostendpos++;
        }
        let [hostname, port] = _parseHostAndPort(url.slice(p, hostendpos));
        if (!hostname.length || hostname.length > TSURL.HostMaxLength) {
            if (!!opts.throwsError) {
                new tserrors_1.TSError(`TSURL: bad hostname in url '${url}'`);
            }
            return null;
        }
        let ipv6Hostname = hostname.startsWith('[') && hostname.endsWith(']');
        if (ipv6Hostname) {
            hostname = hostname.slice(1, hostname.length - 1);
        }
        p = hostendpos;
        let path = '';
        for (let p = hostendpos; p < len; p++) {
            const c = url[p];
            if (c === "'") {
                path += '%27';
            }
            else {
                path += TSURL.AutoEscapeCharSet.has(c) ? encodeURIComponent(c) : c;
            }
        }
        let hash = '';
        const hashIndex = path.indexOf('#');
        if (hashIndex >= 0) {
            hash = path.slice(hashIndex);
            path = path.slice(0, hashIndex);
        }
        let search = '';
        const qmIndex = path.indexOf('?');
        if (qmIndex >= 0) {
            search = path.slice(qmIndex);
            if (search.length === 1) {
                search = '';
            }
            path = path.slice(0, qmIndex);
        }
        if (search.length > 0 && !!opts.refusesParameters) {
            if (!!opts.throwsError) {
                new tserrors_1.TSError(`TSURL: unexpected parameters in url '${url}'`);
            }
            return null;
        }
        if (!path.length) {
            path = '/';
        }
        return new TSURL(protocol, hostname, port, path, hash, auth, search, ipv6Hostname, w3c);
    }
    constructor(protocol, hostname, port, pathname, hash, auth, search, ipv6, w3c) {
        this.toJSON = TSURL.prototype.toString;
        this._auth = auth;
        this._hash = hash;
        this._hostname = hostname;
        this._pathname = pathname;
        this._port = port;
        this._protocol = protocol;
        this._search = search;
        this.searchParams = new URLSearchParams(search);
        this._w3cProtocol = w3c;
        this._ipv6 = ipv6;
    }
    get w3c() { return this._w3cProtocol; }
    get origin() { return this._protocol + '//' + this.host; }
    get host() {
        let ret = this._ipv6 ? '[' + this._hostname + ']' : this._hostname;
        if (this._port.length) {
            ret += ':';
            ret += this._port;
        }
        return ret;
    }
    get href() {
        let auth = this._auth;
        if (auth.length) {
            auth = encodeURIComponent(auth);
            auth = auth.replace(/%3A/i, ':');
            auth += '@';
        }
        let hash = this._hash;
        if (hash.length && hash.charAt(0) !== '#') {
            hash = '#' + hash;
        }
        let search = this._search;
        if (search.length && search.charAt(0) !== '?') {
            search = '?' + search;
        }
        let pathname = this._pathname;
        if (!pathname.startsWith('/')) {
            pathname = '/' + pathname;
        }
        return this._protocol + '//' + auth + this.host + pathname + search + hash;
    }
    get auth() { return this._auth; }
    get hash() { return this._hash; }
    get search() { return this._search; }
    get pathname() { return this._pathname; }
    get hostname() { return this._hostname; }
    set hostname(h) {
        const hn = (0, strings_1.$ascii)((0, strings_1.$ftrim)(h)).toLowerCase();
        if (!hn.length) {
            new tserrors_1.TSError('TSURL: trying to set empry hostname');
        }
        if (hn.length > TSURL.HostMaxLength) {
            new tserrors_1.TSError('TSURL: trying to set a too large hostname');
        }
        this._hostname = hn;
    }
    get port() { return this._port; }
    set port(p) { this._port = p; }
    get protocol() { return this._protocol; }
    set protocol(s) {
        const v = _completeProtocol(s);
        if (v instanceof tserrors_1.TSError) {
            throw v;
        }
        ;
        this._protocol = v;
    }
    isValid(opts = {}) {
        if (!!opts.refusesParameters && this._search.length > 0) {
            return false;
        }
        this._w3cProtocol = TSURL.StandardURLProtocols.has(this.protocol);
        return this._w3cProtocol || _validateProtocol(this.protocol, opts.acceptedProtocols);
    }
    isEqual(other) {
        return this === other || (other instanceof TSURL && this.href === other.href);
    }
    compare(other) {
        if (other === this) {
            return types_1.Same;
        }
        if (!(other instanceof TSURL)) {
            return undefined;
        }
        return this.href < other.href ? types_1.Ascending : (this.href > other.href ? types_1.Descending : types_1.Same);
    }
    toString() { return this.href; }
    toArray() { return [this.href]; }
    clone() {
        return new TSURL(this._protocol, this._hostname, this._port, this._pathname, this._hash, this._auth, this._search, this._ipv6, this._w3cProtocol);
    }
}
exports.TSURL = TSURL;
TSURL.HostMaxLength = 255;
TSURL.AutoEscapeChars = ['{', '}', '|', '\\', '^', '`', "'", '<', '>', '"', '`', ' ', '\r', '\n', '\t'];
TSURL.StandardURLProtocols = new Set(['http:', 'https:', 'file:', 'ftp:', 'ws:', 'wss:']);
TSURL.AutoEscapeCharSet = new Set(TSURL.AutoEscapeChars);
TSURL.EndhostCharSet = new Set(['%', '/', '?', ';', '#', ...TSURL.AutoEscapeChars]);
const __protocolRegex = /^([a-z0-9.+-]+:)/i;
const __protocolCompleteRegex = /^([a-z0-9.+-]+:)$/i;
const __backslashRegex = /\\/g;
const __portRegex = /:[0-9]*$/;
function _normalizeURLString(source) {
    const s = (0, strings_1.$ftrim)(source);
    const index = s.indexOf('?');
    const splitchar = index > -1 && index < s.indexOf('#') ? '?' : '#';
    const parts = s.split(splitchar);
    parts[0] = parts[0].replace(__backslashRegex, '/');
    return parts.join(splitchar);
}
function _completeProtocol(s) {
    if (s instanceof tserrors_1.TSError) {
        return s;
    }
    const p = (0, strings_1.$ftrim)(s).toLowerCase();
    if (!p.length || !p.match(__protocolCompleteRegex)) {
        return _protocolError(p);
    }
    return p;
}
function _findProtocol(s) {
    const m = __protocolRegex.exec(s);
    return (0, commons_1.$ok)(m) ? m[0] : _protocolError(s);
}
function _protocolError(p) { return new tserrors_1.TSError(`TSURL: impossible to set protocol '${p}'`); }
function _parseHostAndPort(s) {
    s = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s)).toLowerCase();
    const m = __portRegex.exec(s);
    if ((0, commons_1.$ok)(m)) {
        const port = m[0];
        return [s.slice(0, s.length - port.length), port.slice(1)];
    }
    return [s, ''];
}
function _validateProtocol(protocol, acceptedProtocols) {
    acceptedProtocols = (0, commons_1.$value)(acceptedProtocols, []);
    let found = acceptedProtocols.find(a => {
        let p = (0, strings_1.$ftrim)(a).toLowerCase();
        if (!p.endsWith(':')) {
            p += ':';
        }
        return p === protocol ? protocol : undefined;
    });
    return (0, commons_1.$ok)(found);
}
//# sourceMappingURL=tsurl.js.map